"""

                 This module is the processes_hub module

The purpose of this module is to direct the information from the process
module into other modules. This module expects some to get some response
or data from thos emodule where its directs the information. The respon-
ses will be saved or plotted depending on the options provided to this
module.

The following are the list of function in this module.

1. process_hub


"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

import multiprocessing
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap
import matplotlib._color_data as mcd

# from BioSANS2020.propagation.propensity import propensity_vec, \
#     propensity_vec_molar
from BioSANS2020.propagation.law_of_localization import law_loc_symbolic
from BioSANS2020.propagation.create_wxmaxima_command import for_wxmaxima
from BioSANS2020.propagation.stochastic.mystiffcle import \
    cle_calculate, cle2_calculate
from BioSANS2020.propagation.stochastic.gillespie_ssa import gillespie_ssa
from BioSANS2020.propagation.stochastic.tau_leaping import tau_leaping
from BioSANS2020.propagation.stochastic.tau_leaping2 import tau_leaping2
from BioSANS2020.propagation.stochastic.mytauleap import sim_tauLeap
# from BioSANS2020.propagation.stochastic.sde_int import SDE_int
from BioSANS2020.propagation.deterministic.euler_mod import euler_int, \
    euler2_int
# from BioSANS2020.propagation.deterministic.lna_approx import *
from BioSANS2020.propagation.deterministic.ode_int import ode_int
from BioSANS2020.propagation.deterministic.lna_function_of_time import \
    lna_non_steady_state, lna_non_steady_state2
from BioSANS2020.propagation.deterministic.runge_kutta4 \
    import rungek4_int, rungek4a_int
from BioSANS2020.propagation.symbolic.lna_approx2 import lna_symbolic2
from BioSANS2020.propagation.symbolic.analytical_sol import analyt_soln
from BioSANS2020.model.param_est.param_estimate import param_estimate

from BioSANS2020.model.fileconvert.convtopotosbml import topo_to_sbml
from BioSANS2020.model.param_est.param_slider import param_ode_int

from BioSANS2020.myglobal import mglobals as globals2
# from BioSANS2020.myglobal import proc_global

from BioSANS2020.gui_functs.draw_figure import draw_figure


def process_hub(
        time_var, sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
        v_volms=1, miter=1, logx=False, logy=False, del_coef=10,
        normalize=False, method="CLE", mix_plot=True,
        save=True, out_fname="", plot_show=True,
        time_unit="time (sec)", vary="", mult_proc=False,
        items=None, vary2="", implicit=False, rfile="",
        exp_data_file=None
):
    """This function  redirect all information from the process to other
    modules and expect the return from those  module for plotting or for
    saving into a file.

    Args:
        time_var (list): list of time points in the simulation
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, -kf1    # negative means to be estimated
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ksn_dict (dict): dictionary of rate constant that appear in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc (dict): dictionary of initial concentration.

            For example;

                {'A': 100.0, 'B': -1.0, 'C': 0.0}
                negative means unknown or for estimation
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stoch_var (numpy.ndarray): stoichiometric matrix. For example

            stoch_var = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        v_volms (float, optional): volume of compartment. Defaults to 1.
        miter (int, optional): Number of iteration or trajectory samples
            for stochastic integration
        logx (bool, optional): If True, the x-axis will be in log scale.
            Defaults to False.
        logy (bool, optional): if True, the y-axis will be in log scale.
            Defaults to False.
        del_coef (float, optional): factor for modifying time steps used
            in the integration/propagation of ODE. Defaults to 10.
        normalize (bool, optional): True  will  be normalized the y axis
            based on max value . Defaults to False.
        method (str, optional): Defaults to "CLE". Any of the option in
            the list of available method keywords is listed below;

            Stochastic (refer to section 10.2.4)

            1.	"CLE"            - Molecules(micro), tau-adaptive
            2.	"CLE2"           - Molecules(micro), cle-fixIntvl
            3.	"Gillespie_"     - Molecules(micro), Direct method
            4.	"Tau-leaping"    - Molecules(micro),
                                   Not swapping with Gillespie
            5.	"Tau-leaping2"   - Molecules(micro),
                                   Swapping with Gillespie
            6.	"Sim-TauLeap"    - Molecules(micro), Simplified,
                                   Swapping with Gillespie

            Deterministic (refer to section 10.2.1)

            7.	"Euler-1"        - Molecules(micro), tau-adaptive-1
            8.	"Euler-2"        - Molar (macro), tau-adaptive-1
            9.	"Euler-3"        - Mole (macro), tau-adaptive-1
            10.	"Euler2-1"	     - Molecules(micro), tau-adaptive-2
            11.	"Euler2-2"       - Molar (macro), tau-adaptive-2
            12.	"Euler2-3"       - Mole (macro), tau-adaptive-2
            13.	"ODE-1"          - Molecules(micro),
                                   using ode_int from scipy
            14.	"ODE-2"          - Molar(macro),
                                   using ode_int from scipy
            15.	"ODE-3"          - Mole(macro), using ode_int from scipy
            16.	"rk4-1"          - Molecules(micro), fix-interval
            17.	"rk4-2"          - Molar(macro), fix-interval
            18.	"rk4-3"          - Mole(macro), fix-interval
            19.	"rk4-1a"         - Molecules(micro), tau-adaptive
            20.	"rk4-2a"         - Molar(macro), tau-adaptive
            21.	"rk4-3a"         - Mole(macro), tau-adaptive

            Linear Noise Approximation (refer to 10.1.2 & 10.2.2)

            22.	"LNA"             - Numeric, values
            23.	"LNA-vs"          - Symbolic, values, Macroscopic
            24.	"LNA-ks"          - Symbolic, f(ks), Macroscopic
            25.	"LNA-xo"          - Symbolic, f(xo), Macroscopic
            26.	"LNA2"            - Symbolic, f(xo,ks), Microscopic
            27.	"LNA3"            - Symbolic, f(xo,ks), Macroscopic
            28.	"LNA(t)"          - COV-time-dependent, Macroscopic
            29.	"LNA2(t)"         - FF-time-dependent, Macroscopic

            Network Localization (refer to 10.1.3)

            30.	"NetLoc1"         - Symbolic, Macroscopic
            31.	"NetLoc2"         - Numeric, Macroscopic

            Parameter estimation (refer to 10.2.3)

            32.	"k_est1"          - MCEM, Macroscopic
            33.	"k_est2"          - MCEM, Microscopic
            34.	"k_est3"          - NM-Diff. Evol., Macroscopic
            35.	"k_est4"          - NM-Diff. Evol., Microscopic
            36.	"k_est5"          - Parameter slider/scanner
            37.	"k_est6"          - Nelder-Mead (NM), Macroscopic
            38.	"k_est7"          - Nelder-Mead (NM), Microscopic
            39.	"k_est8"          - Powell, Macroscopic
            40.	"k_est9"          - Powell, Microscopic
            41.	"k_est10"         - L-BFGS-B, Macroscopic
            42.	"k_est11"         - L-BFGS-B, Microscopic

            Symbolic/Analytical expression of species (refer to 10.1.1)

            43.	"Analyt"          - Pure Symbolic :f(t,xo,k)
            44.	"Analyt-ftx"      - Semi-Symbolic :f(t,xo)
            45.	"SAnalyt"         - Semi-Symbolic :f(t)
            46.	"SAnalyt-ftk"     - Semi-Symbolic :f(t,k)
            47.	"Analyt2"         - Creates commands for wxmaxima
        mix_plot (bool, optional): If True, all species are plotted in
            one plot/figure. Defaults to True.
        save (bool, optional): If True, the resulting trajectory in the
            simulation will be saved as a file. Defaults to True.
        out_fname (str, optional): output filename. Defaults to "".
        plot_show (bool, optional): If True, an image of the plots will
            be created in the directory of the topology file.
        time_unit (str, optional): Defaults to "time (sec)".
        vary (str, optional): Varying initial concentration.
            Defaults to "".
        mult_proc (bool, optional): If True, trajectories will be propa-
            gated on parallel. Defaults to False.
        items (tuple, optional): (canvas, scroll_x, scroll_y).
            Defaults to None.
        vary2 (str, optional): [description]. Defaults to "".
        implicit (bool, optional): True means report in time intervals
            similar to the input time intervals even if actual step is
            more or less. Defaults to False.
        rfile (str): file name of BioSANS topology file.
        exp_data_file ([type], optional): Experimental data file contai-
            ning True or accepted trajectories. Defaults to None.
    """

    stch_temp = []
    nzvar = []
    si_new = []
    slabels = list(sp_comp.keys())  # [xvar for xvar in sp_comp]
    for row in range(stoch_var.shape[0]):
        if np.sum(np.abs(stoch_var[row, :])) != 0 and slabels[row][0] != "-":
            stch_temp.append(list(stoch_var[row, :]))
            nzvar.append(row)
    stoch_var = np.array(stch_temp)
    sp_comp = {slabels[zvar]: sp_comp[slabels[zvar]] for zvar in nzvar}

    slabels = list(sp_comp.keys())
    data = []
    # if len(vary) > 0:
    if vary:
        hold = vary.pop()
        miter = len(vary)
    # if len(vary2) > 0:
    if vary2:
        r_index1, r_index2 = vary2[0]
        kval = vary2[1]
        miter = len(kval)
    if mult_proc:
        pool = multiprocessing.Pool(
            min(miter, round(0.9 * multiprocessing.cpu_count())))
        concn_list = []
        ksns_list = []
        # if len(vary) > 0:
        if vary:
            for j in range(miter):
                concn[hold] = vary[j]
                concn_list.append({xvar: concn[xvar] for xvar in concn})
        else:
            concn_list = [concn for xvar in range(miter)]
        # if len(vary2) > 0:
        if vary2:
            for j in range(miter):
                ksns_list.append({xvar: ksn_dict[xvar][:]
                                  for xvar in ksn_dict})
                if isinstance(r_index1, int):
                    ksns_list[-1][r_index1][r_index2] = kval[j]
                else:
                    for k, _ in enumerate(r_index1):
                        ksns_list[-1][r_index1[k]][r_index2[k]] = kval[k][j]
        else:
            ksns_list = [ksn_dict for xvar in range(miter)]

            # if __name__ == '__main__':
        # if True:  # always true, just use above  command on some OS
        rands = [xvar * np.random.rand() for xvar in range(miter)]
        if method == "Tau-leaping":
            results = [
                pool.apply_async(
                    tau_leaping,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, rands[ih], del_coef,
                        implicit, rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + (slabels,) for result in results]
        elif method == "Tau-leaping2":
            results = [
                pool.apply_async(
                    tau_leaping2,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, rands[ih], del_coef,
                        implicit, rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + (slabels,) for result in results]
        elif method == "Sim-TauLeap":
            results = [
                pool.apply_async(
                    sim_tauLeap,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, rands[ih], del_coef,
                        implicit, rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + (slabels,) for result in results]
        elif method == "CLE":
            results = [
                pool.apply_async(
                    cle_calculate,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, del_coef, rands[ih],
                        implicit, rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + (slabels,) for result in results]
        elif method == "CLE2":
            results = [
                pool.apply_async(
                    cle2_calculate,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, del_coef, rands[ih],
                        rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + (slabels,) for result in results]
        elif method == "Euler-1":
            results = [
                pool.apply_async(
                    euler_int,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, del_coef, False,
                        None, implicit, False, rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + (slabels,) for result in results]
        elif method in ["Euler-2", "Euler-3"]:
            results = [
                pool.apply_async(
                    euler_int,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, del_coef, False,
                        None, implicit, True, rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + (slabels,) for result in results]
        elif method == "LNA":
            save = False
            plot_show = False
            results = [
                pool.apply_async(
                    euler_int,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, del_coef, True,
                        items, False, rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + [slabels] for result in results]
        elif method == "Gillespie_":
            results = [
                pool.apply_async(
                    gillespie_ssa,
                    args=(
                        time_var, sp_comp, ksns_list[ih], concn_list[ih],
                        r_dict, p_dict, stoch_var, rands[ih], implicit,
                        rfile)
                ) for ih in range(miter)
            ]
            data = [result.get() + (slabels,) for result in results]
        else:
            print("Multiprocessing not supported for \
                your method of choice")
        pool.close()
    else:
        rands = [xvar * np.random.rand() for xvar in range(miter)]
        for j in range(miter):
            tnew = []
            r_seed = rands[j]
            # if len(vary) > 0:
            if vary:
                concn[hold] = vary[j]
            # if len(vary2) > 0:
            if vary2:
                if isinstance(r_index1, int):
                    ksn_dict[r_index1][r_index2] = kval[j]
                else:
                    for k, _ in enumerate(r_index1):
                        ksn_dict[r_index1[k]][r_index2[k]] = kval[k][j]
            if method == "CLE":
                tnew, zvar = cle_calculate(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, r_seed, implicit, rfile)
            elif method == "CLE2":
                tnew, zvar = cle2_calculate(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, r_seed, rfile)
            elif method == "Gillespie_":
                tnew, zvar = gillespie_ssa(
                    time_var, sp_comp, ksn_dict, concn, r_dict,
                    p_dict, stoch_var, r_seed, implicit, rfile)
            elif method == "Tau-leaping":
                tnew, zvar = tau_leaping(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, r_seed, del_coef, implicit, rfile)
            elif method == "Tau-leaping2":
                tnew, zvar = tau_leaping2(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, r_seed, del_coef, implicit, rfile)
            elif method == "Sim-TauLeap":
                tnew, zvar = sim_tauLeap(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, r_seed, del_coef, implicit, rfile)
            elif method == "Euler-1":
                tnew, zvar = euler_int(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, False, None, implicit, False,
                    rfile)
            elif method in ["Euler-2", "Euler-3"]:
                tnew, zvar = euler_int(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, False, None, implicit, True,
                    rfile)
            elif method == "Euler2-1":
                tnew, zvar = euler2_int(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, False, None, implicit, False,
                    rfile)
            elif method in ["Euler2-2", "Euler2-3"]:
                tnew, zvar = euler2_int(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, False, None, implicit, True,
                    rfile)
            elif method == "LNA":
                plot_show = False
                save = False
                tnew, zvar = euler_int(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, lna_solve=True, items=items,
                    rfile=rfile)
                tnew, zvar = euler2_int(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, False, None, implicit, True,
                    rfile)
            elif method == "LNA(t)":
                zvar, si_new, tnew = lna_non_steady_state(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, molar=True, rfile=rfile, del_coef=del_coef)
            elif method == "LNA2(t)":
                zvar, si_new, tnew = lna_non_steady_state2(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, molar=True, rfile=rfile, del_coef=del_coef)
            elif method == "LNA-vs":
                plot_show = False
                save = False
                tnew, zvar = lna_symbolic2(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, molar=True, mode="Numeric")
            elif method == "LNA-ks":
                plot_show = False
                save = False
                tnew, zvar = lna_symbolic2(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, molar=True, mode="fofks")
            elif method == "LNA-xo":
                plot_show = False
                save = False
                tnew, zvar = lna_symbolic2(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, molar=True, mode="fofCo")
            elif method == "LNA2":
                plot_show = False
                save = False
                tnew, zvar = lna_symbolic2(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items)
            elif method == "LNA3":
                plot_show = False
                save = False
                tnew, zvar = lna_symbolic2(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, molar=True)
            elif method == "NetLoc1":
                plot_show = False
                save = False
                tnew, zvar = law_loc_symbolic(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, molar=True)
            elif method == "NetLoc2":
                plot_show = False
                save = False
                tnew, zvar = law_loc_symbolic(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, molar=True, numer=True)
            elif method == "ODE-1":
                zvar = ode_int(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, False, rfile)
                tnew = time_var
            elif method in ["ODE-2", "ODE-3"]:
                zvar = ode_int(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, True, rfile)
                tnew = time_var
            # elif method == "Itoint-1":
                # zvar = SDE_int(concn, time_var, sp_comp, ksn_dict,
                               # r_dict, p_dict, stoch_var)
                # tnew = time_var
            # elif method in ["Itoint-2", "Itoint-3"]:
                # zvar = SDE_int(concn, time_var, sp_comp, ksn_dict, r_dict,
                               # p_dict, stoch_var, False)
                # tnew = time_var
            # elif method == "Stratint-1":
                # zvar = SDE_int(
                    # concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    # stoch_var, True, False)
                # tnew = time_var
            # elif method in ["Stratint-2", "Stratint-3"]:
                # zvar = SDE_int(
                    # concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    # stoch_var, False, False)
                # tnew = time_var
            elif method == "rk4-1":
                tnew, zvar = rungek4_int(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, False, del_coef, rfile)
            elif method in ["rk4-2", "rk4-3"]:
                tnew, zvar = rungek4_int(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, True, del_coef, rfile)
            elif method == "rk4-1a":
                tnew, zvar = rungek4a_int(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, False, implicit, rfile)
            elif method in ["rk4-2a", "rk4-3a"]:
                tnew, zvar = rungek4a_int(
                    time_var, sp_comp, ksn_dict, concn, r_dict, p_dict,
                    stoch_var, del_coef, True, implicit, rfile)
            elif method == "k_est1":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=True,
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "k_est2":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, true_data_fil=exp_data_file,
                    rfile=rfile)
            elif method == "k_est3":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=True, mode="DEvol",
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "k_est4":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=False, mode="DEvol",
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "k_est5":
                set_p = [logx, logy, normalize, items]
                zvar = param_ode_int(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, True, rfile, set_p)
                plot_show = False
                save = False
                tnew = time_var
            elif method == "k_est6":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=True, mode="NeldMead",
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "k_est7":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=False, mode="NeldMead",
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "k_est8":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=True, mode="Powell",
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "k_est9":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=False, mode="Powell",
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "k_est10":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=True, mode="L-BFGS-B",
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "k_est11":
                plot_show = False
                save = False
                tnew, zvar = param_estimate(
                    concn, time_var, sp_comp, ksn_dict, r_dict, p_dict,
                    stoch_var, items=items, molar=False, mode="L-BFGS-B",
                    true_data_fil=exp_data_file, rfile=rfile)
            elif method == "Analyt":
                plot_show = False
                save = False
                zvar = analyt_soln(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, rfile=rfile)
            elif method == "Analyt-ftx":
                plot_show = False
                save = False
                zvar = analyt_soln(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, rfile=rfile, mode="ftxo")
            elif method == "SAnalyt":
                plot_show = False
                save = False
                zvar = analyt_soln(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, rfile=rfile, not_semi=False)
            elif method == "SAnalyt-ftk":
                plot_show = False
                save = False
                zvar = analyt_soln(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, rfile=rfile, not_semi=False, mode="ftks")
            elif method == "Analyt2":
                plot_show = False
                save = False
                tnew, zvar = for_wxmaxima(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    items=items, rfile=rfile)
            elif method == "topoTosbml":
                plot_show = False
                save = False
                tnew, zvar = topo_to_sbml(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    v_volms, items=items, molar=False, rfile=rfile)
            elif method == "topoTosbml2":
                plot_show = False
                save = False
                tnew, zvar = topo_to_sbml(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    v_volms, items=items, molar=True, rfile=rfile)
            elif method == "topoTosbml3":
                plot_show = False
                save = False
                tnew, zvar = topo_to_sbml(
                    sp_comp, ksn_dict, concn, r_dict, p_dict, stoch_var,
                    v_volms, items=items, molar=None, rfile=rfile)
            data.append([tnew, zvar, slabels])

    nzvar = []
    reserve_events_words = {
        "t", "time", "status", "status2", "timer", "finish", "delay", "dtime"}
    for row in range(stoch_var.shape[0]):
        key = slabels[row].strip().split("_")[0]
        if key not in reserve_events_words:
            nzvar.append(row)
    sp_comp = {slabels[zvar]: sp_comp[slabels[zvar]] for zvar in nzvar}
    slabels = list(sp_comp.keys())

    # if len(si_new) > 0:
    if si_new:
        slabels = si_new
        sp_comp = si_new
        nzvar = range(len(slabels))

    if save:
        fname = out_fname + "_" + method + ".dat"
        file = open(fname, "w")
        file.write("time\t" + "\t".join(slabels) + "\n")
        for traj in data:
            xvar = traj[0]
            yvar = traj[1]
            xlen = len(xvar)
            for ihv in range(xlen):
                file.write(
                    str(xvar[ihv]) + "\t" + "\t"
                    .join([str(yi) for yi in yvar[ihv][nzvar]]) + "\n")
        file.close()

    if plot_show:

        yaxis_name = "conc"
        if method in ["rk4-1", "rk4-1a", "Euler-1", "Euler2-1", "ODE-1"]:
            yaxis_name = "molecules"
        elif method in ["rk4-2", "rk4-2a", "Euler-2", "Euler2-2", "ODE-2"]:
            yaxis_name = "moles/L"
        elif method in ["rk4-3", "rk4-3a", "Euler-3", "Euler2-3", "ODE-3"]:
            yaxis_name = "moles"
        elif method in ["CLE", "CLE2", "Tau-leaping",
                        "Tau-leaping2", "Sim-TauLeap"]:
            yaxis_name = "molecules"

        sp_len = len(sp_comp)
        if sp_len <= 10:
            col = ['C' + str(i) for i in range(10)]
        elif sp_len > 10 and sp_len <= 40:
            col = list(get_cmap("tab20").colors) \
                + list(get_cmap("tab20b").colors)
        else:
            col = list(mcd.CSS4_COLORS)
        if mix_plot:
            plt.figure(figsize=(9.68, 3.95))
            plt.xlabel(time_unit)
            plt.ylabel(yaxis_name)
            # if len(si_new) > 0:
            if si_new:
                plt.ylabel("cov" if si_new[0][0:2] != "FF" else "FF")
            lines = []
            if logx:
                plt.xscale('log')
            if logy:
                plt.yscale('log')

            for j in range(miter):
                for i in nzvar:
                    if normalize:
                        line = plt.plot(
                            data[j][0],
                            data[j][1][:, i] / (np.max(data[j][1][:, i])
                                                + 1.0e-30),
                            color=col[i])
                    else:
                        line = plt.plot(data[j][0], data[j]
                                        [1][:, i], color=col[i])
            if len(slabels) <= 10:
                plt.legend(slabels)
            else:
                plt.legend(slabels, fontsize='xx-small')
            plt.tight_layout()
            plt.savefig(out_fname + "_" + method + ".jpg")
            fig = plt.gcf()
            lines.append(line)
            globals2.PLOTTED.append([plt.gca(), fig, lines])
            # fig_canvas_agg = draw_figure(items, fig)
            draw_figure(items, fig)
            # plt.close()
        else:
            lines = []
            s_si = []
            for i in nzvar:
                s_si.append(slabels[i])
                plt.figure(figsize=(9.68, 3.95))
                plt.xlabel(time_unit)
                plt.ylabel(yaxis_name)
                # if len(si_new) > 0:
                if si_new:
                    plt.ylabel("cov" if si_new[0][0:2] != "FF" else "FF")
                if logx:
                    plt.xscale('log')
                if logy:
                    plt.yscale('log')
                for j in range(miter):
                    if normalize:
                        line = plt.plot(
                            data[j][0],
                            data[j][1][:, i] / (np.max(data[j][1][:, i])
                                                + 1.0e-30),
                            color=col[i])
                    else:
                        line = plt.plot(data[j][0], data[j]
                                        [1][:, i], color=col[i])
                    lines.append(line)
                plt.legend([slabels[i]])
                plt.tight_layout()
                plt.savefig(out_fname + "_" + method + "_" + str(i) + ".jpg")
                fig = plt.gcf()
                globals2.PLOTTED.append([plt.gca(), fig, lines])
                # fig_canvas_agg = draw_figure(items, fig)
                draw_figure(items, fig)
                # plt.close()
    return data



"""
                       This is the scrollable_text module

This module controls how text area are placed inside the canvas for
displaying output of computation, equations, etc.

The following is the list of function for this module:

1. save_file
2. tab
3. delete_this
4. prepare_scroll_text

"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

from tkinter import Text, INSERT, END, Scrollbar, Frame, Button
from tkinter import filedialog
from sys import platform

from BioSANS2020.myglobal import mglobals as globals2


def save_file(_, text):
    """Opens a file dialog box for saving the constents of the text area"""
    # global file_name
    file = filedialog.asksaveasfile(mode='w', defaultextension=".txt")
    if file is None:
        return -1
    file.write(text.get("0.0", END))
    file.close()
    return "break"


def tab(_, text):
    """insers tab or four spaces in the text area"""
    text.insert(INSERT, " " * 4)
    return "break"


def canvas_update_widgets(_, canvas):
    """This function update the canvas upon adding new objects by moving
    objects to ensure they fit nicely. The arrangement is that the last
    object added appear first (latest goes on top).
    Args:
        canvas : the canvas object"""
    #R = canvas._root().winfo_height()
    #root = canvas._root()
    #width = root.winfo_screenwidth()
    #height = root.winfo_screenheight()
    # print(canvas.winfo_children())
    canvas_height = canvas.winfo_height()
    canvas_width = canvas.winfo_width()
    obj_canv = canvas.find_all()
    obj_len = len(obj_canv)
    ind = 0
    for xvar in obj_canv:
        canvas.itemconfig(xvar, height=canvas_height, width=canvas_width - 5)
        # xx1, yy1, xx2, yy2 = canvas.bbox(xvar)
        _, yy1, _, _ = canvas.bbox(xvar)
        canvas.move(xvar, 0, (obj_len - ind) * canvas_height - yy1 + 3)
        ind = ind + 1
    canvas.configure(scrollregion=canvas.bbox("all"))
    return "break"


def delete_this(frame, canvas):
    """This function removes objects from the canvas and reorder objects"""
    canvas.delete(frame)
    canvas_update_widgets(None, canvas)


def prepare_scroll_text(items):
    """This function prepares the scrollable text area by creating a
    frame as a child of the canvas or items[0]. Scroll capabilities are
    added and a Text area is created as a child of the frame.
    Args:
        items : 3 item list of [canvas, scroll_x, scroll_y]
    Returns:
        text : the text area object where text can be inserted
    """
    # canvas, scroll_x, scroll_y = items
    canvas, _, scroll_y = items
    frame = Frame(canvas, height=455, width=940, borderwidth=10, bd=0)

    hscroll = Scrollbar(frame, orient="horizontal")
    hscroll.pack(side="bottom", fill="x")
    vscroll = Scrollbar(frame)
    vscroll.pack(side="right", fill="y")

    font_val = 11 if platform.lower() != "darwin" else 15
    text = Text(frame, width=int(106 * canvas.winfo_width() / 972), height=25,
                fg='blue', font=("Courier New", font_val, "bold"), wrap="none")
    text.configure(xscrollcommand=hscroll.set, yscrollcommand=vscroll.set)
    text.pack(side="top", fill="both", expand=True)

    frame.pack(side="top", fill="x", expand=True)

    vscroll.config(command=text.yview)
    hscroll.config(command=text.xview)

    fframe = canvas.create_window(
        0, 450 * globals2.PLOT_I, anchor='nw', window=frame)
    bttn = Button(frame, text=" X ", fg='red', highlightcolor='blue',
                  bg='white', height=1, relief='raised',
                  command=lambda: delete_this(fframe, canvas))
    bttn.place(rely=0.0, relx=1.0, x=-15, y=0, anchor="ne")

    # canvas.configure(yscrollcommand=scroll_y.set,
    #                  xscrollcommand=scroll_x.set)
    canvas.configure(yscrollcommand=scroll_y.set)
    canvas.configure(scrollregion=canvas.bbox("all"))
    globals2.PLOT_I = globals2.PLOT_I + 1
    canvas_update_widgets(None, canvas)

    text.bind("<Button-2>", lambda e: save_file(e, text))
    text.bind("<Tab>", lambda e: tab(e, text))
    # text.bind("<Button-3>",lambda e: print("kkkk"))
    canvas.bind("<Configure>", lambda e: canvas_update_widgets(e, canvas))
    # canvas.bind("<Configure>", lambda e: [ canvas.itemconfig(xx,width= \
    #     canvas.winfo_width()) for xx in canvas.find_all() ])
    return text



"""

This module is for initial test

"""


from numpy import linspace, sin, cos, pi, vstack, random
label = "Thank you for using BioSANS".split()
theta = linspace(0, 2 * pi, 100)
x = 16 * (sin(theta) ** 3)
tup = (x,)
for i in range(5):
    y = 13 * cos(theta)-5*cos(2*theta)-2*cos(3*theta)-cos(4*theta) \
        + 5*random.rand()
    tup = tup + (y,)
data = [vstack(tup).T]



"""

                   This module is the BioSANS module

The purpose of this module  is to provide a graphical user interface and
to facilitate  the transfer of information from BioSANS topology file to
other modules.


The following are the list of functions in this module

1. load_data
2. show_file_dir
3. create_file
4. extract_ode
5. sbml_to_topo2
6. save_file
7. runpy_file
8. run_ssl
9. load_data2
10. tload_data2
11. delete_this
12. canvas_update_widgets
13. load_image
14. eval2
15. dict_trans
16. convert
17. range_trans
18. range_prep
19. mrun_propagation
20. tprocess
21. analysis_case
22. plot_traj_d2
23. param_set


"""

import os
import tkinter as gui
from tkinter import ttk  # , Button, filedialog
import time
import webbrowser
from sys import platform as PLATFORM, executable as sys_executable
from math import ceil as Myceil
from pathlib import Path
from datetime import datetime
import threading
from PIL import Image as Image2, ImageTk
from numpy import array as np_array

# from queue import Queue


from BioSANS2020.test_data import test_data2 as test_data2
from BioSANS2020.myglobal import mglobals as globals2
from BioSANS2020.myglobal import proc_global
from BioSANS2020.gui_functs.prepare_canvas import prepare_frame_for_plot
from BioSANS2020.prepcodes.process import process
from BioSANS2020.analysis.plotting.plot_traj \
    import plot_traj, plot_traj2
from BioSANS2020.analysis.numeric.transform_data import calc_cross_corr, \
    calc_covariance, fano_factor, prob_density_calc, \
    prob_density_calc2, prob_density_calc3, ave_traj_calc
from BioSANS2020.model.fileconvert.process_sbml \
    import process_sbml as sbml_to_topo
from BioSANS2020.model.ode_parse import ode_extract

from BioSANS2020.model import topology_view
from BioSANS2020.model.new_file import new_file

if PLATFORM == "win32":
    from subprocess import Popen, CREATE_NEW_CONSOLE
elif PLATFORM == "darwin":
    try:
        from applescript import tell as my_tell_us
    except BaseException:
        pass
    from subprocess import check_output, call as my_call_us
elif PLATFORM == "linux":
    pass
else:
    from subprocess import Popen

try:
    import tempfile
    TEMPORARY_FOLDER = str(tempfile.gettempdir()).replace(
        "\\", "/") + "/BioSANS_temporary_folder"
except BaseException:
    TEMPORARY_FOLDER = "BioSANS_temporary_folder"

globals2.init(globals2)
if __name__ == '__main__':
    proc_global.init(proc_global)

TOP = gui.Tk()
TOP.title("BioSANS 1.0")
TOP.geometry("1005x550")
# TOP.resizable(False, False)

HEADER = gui.Label(TOP, text="BioSANS")
HEADER.configure(
    bg="green",
    fg="white",
    height=1,
    # width = 1005,
    font="Helvetica 18 bold italic"
)
HEADER.pack(fill="x")

FRAME = gui.Frame(TOP)
FRAME.configure(
    bg="light cyan",
    borderwidth=2,
    height=500,
    width=1005
)
FRAME.pack(fill="both", expand=True)

FOOTER = gui.Label(
    TOP, text="Biological Symbolic and Numeric Simulation Algorithms")
FOOTER.configure(
    bg="green",
    fg="white",
    # width = 1005,
    font="Helvetica 10 bold italic",
    anchor='w'
)
FOOTER.pack(fill="x")

FILE_NAME = {"topology": TEMPORARY_FOLDER, "current_folder": None}


def load_data(itups):
    """This fuction reads topology file and display the contents in a
    text area.

    Args:
        itups (tuple): (canvas, scroll_x, scroll_y)
    """
    # global FILE_NAME
    file = gui.filedialog.askopenfilename(title="Select file")
    FILE_NAME["topology"] = file
    FILE_NAME["current_folder"] = file
    globals2.TO_CONVERT = file
    if os.path.isfile(file):
        FILE_NAME['last_open'] = topology_view.view_topo(file, itups)


def show_file_dir(path):
    """This function opens the current working directory

    Args:
        path (str): directory path
    """
    # global PLATFORM
    if PLATFORM == "win32":
        os.startfile(os.path.dirname(path))
    elif PLATFORM == "darwin":
        my_call_us('open', os.path.dirname(path))
    elif PLATFORM == "linux":
        my_call_us('xdg-open', os.path.dirname(path))
    else:
        webbrowser.open(os.path.dirname(path))


def create_file(itups, ftype):
    """This function creates a temporary file in a temporary directory
    where new topology / files can be placed.

    Args:
        itups (tuple): (canvas, scroll_x, scroll_y)
        ftype (int): 1 for BioSANS topology file, 2 for ODE file
    """
    # global FILE_NAME, TEMPORARY_FOLDER
    try:
        os.mkdir(TEMPORARY_FOLDER, 0o777)
    except BaseException:
        for item in Path(TEMPORARY_FOLDER).iterdir():
            if item.is_dir():
                pass
            else:
                item.unlink()

    FILE_NAME['last_open'] = new_file(itups)
    FILE_NAME["topology"] = TEMPORARY_FOLDER + "/temp.txt"
    if ftype == 1:
        FILE_NAME['last_open'].insert('insert', "FUNCTION_DEFINITIONS:\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert(
            'insert', "#REACTIONS, Volume = 1, tend = 100, steps = 100, "
            + "FileUnit = molar\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "@CONCENTRATION\n")
    elif ftype == 2:
        FILE_NAME['last_open'].insert('insert', "ODE_DECLARATIONS:\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "INI_CONCENTRATIONS:\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "\n")
        FILE_NAME['last_open'].insert('insert', "RATE_CONSTANTS:\n")


def extract_ode(itups):
    """This function extracts topology from ODE file format

    Args:
        itups (tuple): (canvas, scroll_x, scroll_y)
    """
    # global FILE_NAME
    FILE_NAME['last_open'] = ode_extract.odedxdt_to_topo(
        FILE_NAME["topology"], itups)
    FILE_NAME["topology"] = FILE_NAME["topology"] + ".TOP"


def sbml_to_topo2(tocon, itups):
    """This function helps  in the conversion  of SBML files  to BioSANS
    topology files.

    Args:
        tocon (str): sbml file name to convert
        itups (tuple): (canvas, scroll_x, scroll_y)
    """
    FILE_NAME["topology"] = sbml_to_topo(tocon)
    FILE_NAME['last_open'] = topology_view.view_topo(
        FILE_NAME["topology"], itups)


def save_file():
    """This function saves the content of the last opened text area into
    a file.
    """
    # global FILE_NAME
    file = gui.filedialog.asksaveasfile(mode='w', defaultextension=".txt")
    FILE_NAME["topology"] = file.name
    FILE_NAME["current_folder"] = file.name
    if file is None:
        return
    file.write(FILE_NAME['last_open'].get("0.0", "end"))
    file.close()


def runpy_file():
    """This module write the contents of the text area into a python
    file and run the file as a python script."""
    # global FILE_NAME, PIPE
    with open(FILE_NAME["topology"], "w") as ffvar:
        ffvar.write(FILE_NAME['last_open'].get("0.0", "end"))
        ffvar.write("\ninput('Press enter to exit:')")
    if PLATFORM == "win32":
        Popen([sys_executable, FILE_NAME["topology"]],
              creationflags=CREATE_NEW_CONSOLE)
    elif PLATFORM == "darwin":
        my_tell_us.app("Terminal", 'do script "' +
                       str(sys_executable) + " " + FILE_NAME["topology"] + '"')
    elif PLATFORM == "linux":
        os.system("gnome-terminal -x python3 " + FILE_NAME["topology"])
    else:
        Popen(str(sys_executable) + " " + FILE_NAME["topology"], shell=True)


def run_ssl():
    """This function initiates BioSSL or the BioSANS structured
    simulation language."""
    if PLATFORM == "win32":
        Popen([sys_executable, "-m", "BioSANS2020.BioSSL"],
              creationflags=CREATE_NEW_CONSOLE)
    elif PLATFORM == "darwin":
        pip_show_out = check_output(['pip3', 'show', 'BioSANS2020-efajiculay'])
        pip_show_out = str(pip_show_out).split("\\n")
        install_dir = ""
        for row in pip_show_out:
            line = row.split(":")
            if line[0].strip() == "Location":
                install_dir = (
                    "".join(line[1:])
                    .strip("\\r\\n").replace("c\\", "c:/")
                    .replace("\\", "/").replace("//", "/"))
        if install_dir != "":
            install_dir = str(install_dir)
            my_tell_us.app("Terminal", 'do script "' + str(sys_executable) +
                           " " + install_dir + "/BioSANS2020/BioSSL.py" + '"')
    elif PLATFORM == "linux":
        os.system("gnome-terminal -x python3 " +
                  os.path.join(os.getcwd(), "BioSSL.py"))
    else:
        Popen(str(sys_executable) + " " +
              os.path.join(os.getcwd(), "BioSSL.py"), shell=True)


def load_data2(plot=False, def_data=None):
    """This function loads data for numerical processing or for plotting

    Args:
        plot (bool, optional): if True, loaded trajectory is plotted.
        Defaults to False.
    """
    # t_o = time.time()
    global CURRENT_DATA
    if not def_data:
        file = gui.filedialog.askopenfilename(title="Select file")
        FILE_NAME["trajectory"] = file
        FILE_NAME["current_folder"] = file
        with open(FILE_NAME["trajectory"], "r") as fvar:
            data = []
            ddvar = []
            row1 = str(fvar.readline()).strip()
            slabels = row1.split("\t")[1:]
            for row in fvar:
                cols = [float(xvar) for xvar in row.split("\t")]
                if cols[0] == 0.0 and ddvar:
                    data.append(np_array(ddvar))
                    ddvar = []
                ddvar.append(cols)
            data.append(np_array(ddvar))
    else:
        data, slabels = def_data
    if plot:
        plot_traj(data, slabels, ITUPS, globals2.PLOTTED, mix_plot=True,
                  logx=False, logy=False, normalize=False)
    # print(data[0], "\n\n")
    # print(data[1], "\n\n")
    # print(data[2], "\n")
    CURRENT_DATA = (data, slabels)
    if not def_data:
        gui.messagebox.showinfo("showinfo", "Trajectory loaded succesfully")
    # print(time.time()-t_o)


def tload_data2(plot=False):
    """This function starts a thread that handles loading data.

    Args:
        plot (bool, optional): if True, loaded trajectory is plotted.
        Defaults to False.
    """
    if __name__ == '__main__':
        tvar = threading.Thread(target=load_data2, args=(plot,), daemon=False)
        tvar.start()


def delete_this(frame, canvas):
    """This function delete an object in the canvas

    Args:
        frame (tkinter.Frame): frame or other objects
        canvas (tkinter.Canvas): canvas object
    """
    canvas.delete(frame)
    canvas_update_widgets(None, canvas)


def canvas_update_widgets(_, canvas):
    """This function rearranged the elements in the canvas.

    Args:
        _ (None): Not needed
        canvas (tkinter.Canvas): canvas object]

    Returns:
        str: "break"
    """
    # R = canvas._root().winfo_height()
    # root = canvas._root()
    # width = root.winfo_screenwidth()
    # height = root.winfo_screenheight()
    # print(canvas.winfo_children())
    cheigh = canvas.winfo_height()
    cwidth = canvas.winfo_width()
    obj_can = canvas.find_all()
    obj_len = len(obj_can)
    ind = 0
    for xvar in obj_can:
        canvas.itemconfig(xvar, height=cheigh, width=cwidth - 5)
        # x_1, y_1, x_2, y_2 = canvas.bbox( xvar)
        _, y_1, _, _ = canvas.bbox(xvar)
        canvas.move(xvar, 0, (obj_len - ind) * cheigh - y_1 + 3)
        ind = ind + 1
    canvas.configure(scrollregion=canvas.bbox("all"))
    return "break"


def load_image(wdata=False):
    """This function load image into a canvas and display in BioSANS.

    Args:
        wdata (bool, optional): if True, data will also be loaded in the
            memory. Defaults to False.
    """
    t_o = time.time()
    global CURRENT_DATA
    canvas, scroll_x, scroll_y = ITUPS
    file = gui.filedialog.askopenfilename(title="Select file")
    FILE_NAME["current_folder"] = file
    load = Image2.open(file)
    render = ImageTk.PhotoImage(load)
    img = gui.Label(canvas, image=render)
    img.image = render

    fframe = canvas.create_window(
        0, 426 * globals2.PLOT_I, anchor='nw', window=img)
    but_b = gui.Button(img, text=" X ", fg='red', highlightcolor='blue',
                       bg='white', height=1, relief='raised',
                       command=lambda: delete_this(fframe, canvas))
    but_b.place(rely=0.0, relx=1.0, x=-15, y=0, anchor="ne")

    canvas.configure(yscrollcommand=scroll_y.set, xscrollcommand=scroll_x.set)
    canvas.configure(scrollregion=canvas.bbox("all"))
    canvas.bind("<Configure>", lambda e: canvas_update_widgets(e, canvas))
    globals2.PLOT_I = globals2.PLOT_I + 1

    if wdata:
        file = str(file).replace("jpg", "dat").replace("png", "dat")
        with open(file, "r") as ffile:
            data = []
            ddvar = []
            row1 = str(ffile.readline()).strip()
            slabels = row1.split("\t")[1:]
            for row in ffile:
                cols = [float(xvar) for xvar in row.split("\t")]
                if cols[0] == 0.0 and ddvar:
                    data.append(np_array(ddvar))
                    ddvar = []
                ddvar.append(cols)
            data.append(np_array(ddvar))
        CURRENT_DATA = (data, slabels)
    print(time.time() - t_o)
    canvas_update_widgets(None, canvas)


def eval2(xvar):
    """This function evaluates expression

    Args:
        xvar (str): expression

    Returns:
        str: evaluated expression
    """
    try:
        return eval(xvar)
    except BaseException:
        try:
            par = str(xvar).lower().capitalize()
            return eval(par)
        except BaseException:
            return str(xvar)


CURRENT_DATA = None


def dict_trans(x_1):
    """This function creates a dictionary from a list assignment
    Args:
        x_1 (str): string of symbols and assinged values.

    Returns:
        dict: a dictionary from a list assignment
    """
    x_2 = x_1.split(",")
    x_3 = {}
    try:
        for xvar in x_2:
            rvar = xvar.split("=")
            x_3[rvar[0].strip()] = float(rvar[1])
    except BaseException:
        pass
    return x_3


def convert(xvar, con):
    """This function converts a variable into con data type.

    Args:
        xvar (str): string
        con (con): new data type con

    Returns:
        con: xvar equivalent in con
    """
    try:
        return con(xvar)
    except BaseException:
        return xvar


def range_trans(x_1):
    """This function converts the x_1 comma concatenated string into a
    list and put the first element as the last element.

    Args:
        x_1 (str): comma concatenated string

    Returns:
        list: the first element goes last now
    """
    x_3 = []
    x_2 = x_1.split(",")
    if len(x_2) > 1:
        x_3 = x_2[1:]
        x_3.append(x_2[0])
    return x_3


def range_prep(x_1):
    """This function process some string input and converts them to a
    list which is used as a range on other fuctions.

    Args:
        x_1 (str): comma concatenated string

    Returns:
        list: list of float or list of list and floats
    """
    x_3 = []
    x_2 = x_1.split(",")
    if len(x_2) > 1:
        x_2 = [convert(x_2[xvar], float) for xvar in range(len(x_2))]
        c_c = x_2[0].lower().split("f")
        r_2 = 0
        if len(c_c) < 2:
            c_c = x_2[0].lower().split("b")
            r_2 = 1
        r_1 = int(c_c[1]) - 1
        c_c = ",".join([str(xvar) for xvar in x_2[1:]])
        x_3 = list(eval(c_c))
        return [[r_1, r_2], x_3]
    return x_1


def mrun_propagation(par, entry_list, defs2):
    """Ths function grabs the values from defs2 whcih serves as the set
    of input for the tprocess function.

    Args:
        par (tkinter.Toplevel): top level container
        entry_list (list): list of tkinter.Entry or tkinter.OptionMenu
        defs2 (list): values
    """
    # global CURRENT_DATA
    try:
        del FILE_NAME["trajectory"]
    except BaseException:
        pass

    defs = []
    for i in range(len(entry_list)):
        try:
            val = eval2(entry_list[i].get())
            defs.append(val)
        except BaseException:
            val = eval2(defs2[i].get())
            defs.append(val)
    # defs[15] = dict_trans(entry_list[15].get())
    defs[16] = range_trans(entry_list[16].get())
    defs[17] = range_prep(entry_list[17].get())
    if not defs[9] in [
            "Analyt", "SAnalyt", "Analyt-ftx", "SAnalyt-ftk", "k_est1",
            "k_est2", "k_est3", "k_est4", "k_est5", "k_est6", "k_est7",
            "k_est8", "k_est9", "k_est10", "k_est11", "NetLoc1", "NetLoc2"]:
        with open(defs[13] + "_" + defs[9] + "_params.dat", "w") as fvar:
            fvar.write("\n".join([str(xvar) for xvar in defs]))
    if defs[9] in [
            "k_est1", "k_est2", "k_est3", "k_est4", "k_est6", "k_est7",
            "k_est8", "k_est9", "k_est10", "k_est11", "LNA2", "LNA3", "Analyt",
            "Analyt-ftx", "SAnalyt", "SAnalyt-ftk", "Analyt2", "topoTosbml",
            "topoTosbml2", "topoTosbml3", "LNA-vs", "LNA-ks", "LNA-xo",
            "NetLoc1", "NetLoc2"]:
        par.destroy()
    defs.append(ITUPS)
    tprocess(defs)


SUPER_THREAD_RUN = None


def tprocess(defs):
    """This function creates a thread and sed defs to the
    BioSANS2020.prepcodes.process function.

    Args:
        defs (list): inputs for process fucntion
    """
    global SUPER_THREAD_RUN
    if defs[9] == "k_est5":
        process(*defs)
    else:
        if __name__ == '__main__':
            SUPER_THREAD_RUN = 1
            tvar = threading.Thread(
                target=lambda: process(*defs), daemon=False)
            tvar.start()


def analysis_case(ana_case, itups):
    """This function  redirects input to  the corresponding numerical or
    plotting processes.

    Args:
        ana_case (str): type of analysis
        itups (tuple): (canvas, scroll_x, scroll_y)

    Returns:
        np.ndarray : numerical values or None
    """
    global FILE_NAME
    if "trajectory" not in FILE_NAME:
        gui.messagebox.showinfo(
            "Trajectory not loaded yet", "Please load the trajectory. BioSANS \
            save it into a file during your last run.")
        try:
            load_data2(False)
        except BaseException:
            try:
                del FILE_NAME["trajectory"]
            except BaseException:
                pass
            return None
        # data, slabels = CURRENT_DATA

    if ana_case == "cov":
        return calc_covariance(CURRENT_DATA, itups)
    if ana_case == "fanoF":
        return fano_factor(CURRENT_DATA, itups)
    if ana_case == "corr":
        return calc_cross_corr(CURRENT_DATA, itups)
    if ana_case == "pdens1":
        return prob_density_calc(CURRENT_DATA, itups)
    if ana_case == "pdens2":
        return prob_density_calc2(CURRENT_DATA, itups)
    if ana_case == "pdens3":
        return prob_density_calc3(CURRENT_DATA, itups)
    if ana_case == "avetrj":
        return ave_traj_calc(CURRENT_DATA, itups)
    if ana_case == "phaseP":
        return plot_traj_d(CURRENT_DATA, itups)
    if ana_case == "plotD":
        return plot_traj_d2(CURRENT_DATA, itups)
    return None


def plot_traj_d(current_data, itups):
    """This function plot the trajectory data stored as current data.
    The plot can be the phase portrait.

    Args:
        current_data (np.ndarray or list): loaded data
        itups (tuple): (canvas, scroll_x, scroll_y)
    """
    # global SUPER_THREAD_RUN
    if SUPER_THREAD_RUN == 1:
        gui.messagebox.showinfo(
            "Warning: Unsafe thread",
            "If you want to have a 3D Phase portrait, restart BioSANS and \
            load trajectory. For 2D phase portrait, just continue")
    try:
        data, slabels = current_data
        par = gui.Toplevel()
        par.resizable(False, False)
        par.wm_title("Plot settings")

        entry_lst = [
            gui.Label(par, text="choose x-axis", fg="blue"),
            gui.Label(par, text="choose y-axis", fg="blue"),
            gui.Label(par, text="choose z-axis", fg="blue"),
            gui.Label(par, text="choose step-range", fg="blue")
        ]
        for i in range(4):
            entry_lst[i].grid(row=i, column=0, sticky=gui.W, pady=2)
        optsel = [[par, gui.StringVar(), "None", "time"] +
                  slabels for i in range(3)]
        opt_sel_var = [optsel[i][1] for i in range(3)]
        for i in range(3):
            opt_sel_var[i].set("None")
        entry_list = [gui.OptionMenu(*optsel[i]) for i in range(3)]
        for i in range(3):
            entry_list[i].config(width=14)
        entry_list.append(gui.Entry(par, bd=5))
        for i in range(4):
            entry_list[i].grid(row=i, column=1, sticky=gui.W, pady=2)
        entry_list[-1].insert(gui.END, "0:-1")

        but_b1 = ttk.Button(par, text="PLOT", command=lambda: plot_traj2(
            data, slabels, itups, globals2.PLOTTED, logx=False, logy=False,
            normalize=False, xlabel=opt_sel_var[0].get(),
            ylabel=opt_sel_var[1].get(), zlabel=opt_sel_var[2].get(),
            trange=entry_list[-1].get()))
        but_b1.grid(row=5, column=0, sticky=gui.W, pady=2)
    except BaseException:
        gui.messagebox.showinfo(
            "Trajectory not loaded yet",
            "Please load the trajectory. BioSANS save it into a file during \
            your last run.")


def get_checked(el_1, slabels):
    """This function returns the components name with a check from el_1
    check boxes list.

    Args:
        el_1 (list): list of values
        slabels (list):components or species labels

    Returns:
        [type]: [description]
    """
    check_si = []
    for i, _ in enumerate(el_1):
        key = el_1[i].get()
        if key != "0":
            check_si.append(slabels.index(key))
    return check_si


def plot_traj_d2(current_data, itups):
    """This function is another plotting function.

    Args:
        current_data (np.ndarray or list): loaded data
        itups (tuple): (canvas, scroll_x, scroll_y)
    """
    try:
        data, slabels = current_data
        pard = gui.Toplevel()
        pard.resizable(True, True)
        pard.wm_title("Plot species")
        pard.maxsize(width=300, height=700)
        canvas, scroll_x, scroll_y = prepare_frame_for_plot(
            pard, width=200, height=200)
        par = gui.Frame(canvas, width=200, height=int(len(slabels) * 150))
        par.pack(side="left", fill="both", expand=True)
        pard.configure(
            bg="light blue",
            borderwidth=2,
        )
        canvas.configure(
            bg="light blue",
            borderwidth=2,
        )

        el_s = len(slabels)
        el_1 = [gui.StringVar() for i in range(el_s)]
        for i in range(el_s):
            el_1[i].set("0")
        entry_lst = [
            gui.Checkbutton(
                par, text=slabels[i], variable=el_1[i], onvalue=slabels[i],
                offvalue="0") for i in range(el_s)]

        for i in range(Myceil(el_s / 2)):
            entry_lst[i].grid(row=i, column=0, sticky=gui.W, pady=2)
        for i in range(Myceil(el_s / 2), el_s):
            entry_lst[i].grid(
                row=i - Myceil(el_s / 2), column=1, sticky=gui.W, pady=2)

        canvas.create_window(0, 0, anchor='n', window=par)
        canvas.configure(yscrollcommand=scroll_y.set,
                         xscrollcommand=scroll_x.set)
        canvas.configure(scrollregion=canvas.bbox("all"))

        but_b1 = ttk.Button(
            pard, text="PLOT",
            command=lambda: plot_traj(
                data, slabels, itups, globals2.PLOTTED, mix_plot=True,
                logx=False, logy=False, normalize=False,
                si_ticked=get_checked(el_1, slabels)))
        but_b1.pack(side="bottom", fill="x")
    except BaseException:
        gui.messagebox.showinfo(
            "Trajectory not loaded yet",
            "Please load the trajectory. BioSANS save it into a file \
                during your last run.")


def param_set(method):
    """This function opens the parameter setting dialof box and grab the
    user custom settings.

    Args:
        method (str): Defaults to "CLE". Any of the option in
            the list of available method keywords is listed below;

            Stochastic (refer to section 10.2.4)

            1.    "CLE"            - Molecules(micro), tau-adaptive
            2.    "CLE2"           - Molecules(micro), cle-fixIntvl
            3.    "Gillespie_"     - Molecules(micro), Direct method
            4.    "Tau-leaping"    - Molecules(micro),
                                   Not swapping with Gillespie
            5.    "Tau-leaping2"   - Molecules(micro),
                                   Swapping with Gillespie
            6.    "Sim-TauLeap"    - Molecules(micro), Simplified,
                                   Swapping with Gillespie

            Deterministic (refer to section 10.2.1)

            7.    "Euler-1"        - Molecules(micro), tau-adaptive-1
            8.    "Euler-2"        - Molar (macro), tau-adaptive-1
            9.    "Euler-3"        - Mole (macro), tau-adaptive-1
            10.    "Euler2-1"         - Molecules(micro), tau-adaptive-2
            11.    "Euler2-2"       - Molar (macro), tau-adaptive-2
            12.    "Euler2-3"       - Mole (macro), tau-adaptive-2
            13.    "ODE-1"          - Molecules(micro),
                                   using ode_int from scipy
            14.    "ODE-2"          - Molar(macro),
                                   using ode_int from scipy
            15.    "ODE-3"          - Mole(macro), using ode_int from scipy
            16.    "rk4-1"          - Molecules(micro), fix-interval
            17.    "rk4-2"          - Molar(macro), fix-interval
            18.    "rk4-3"          - Mole(macro), fix-interval
            19.    "rk4-1a"         - Molecules(micro), tau-adaptive
            20.    "rk4-2a"         - Molar(macro), tau-adaptive
            21.    "rk4-3a"         - Mole(macro), tau-adaptive

            Linear Noise Approximation (refer to 10.1.2 & 10.2.2)

            22.    "LNA"             - Numeric, values
            23.    "LNA-vs"          - Symbolic, values, Macroscopic
            24.    "LNA-ks"          - Symbolic, f(ks), Macroscopic
            25.    "LNA-xo"          - Symbolic, f(xo), Macroscopic
            26.    "LNA2"            - Symbolic, f(xo,ks), Microscopic
            27.    "LNA3"            - Symbolic, f(xo,ks), Macroscopic
            28.    "LNA(t)"          - COV-time-dependent, Macroscopic
            29.    "LNA2(t)"         - FF-time-dependent, Macroscopic

            Network Localization (refer to 10.1.3)

            30.    "NetLoc1"         - Symbolic, Macroscopic
            31.    "NetLoc2"         - Numeric, Macroscopic

            Parameter estimation (refer to 10.2.3)

            32.    "k_est1"          - MCEM, Macroscopic
            33.    "k_est2"          - MCEM, Microscopic
            34.    "k_est3"          - NM-Diff. Evol., Macroscopic
            35.    "k_est4"          - NM-Diff. Evol., Microscopic
            36.    "k_est5"          - Parameter slider/scanner
            37.    "k_est6"          - Nelder-Mead (NM), Macroscopic
            38.    "k_est7"          - Nelder-Mead (NM), Microscopic
            39.    "k_est8"          - Powell, Macroscopic
            40.    "k_est9"          - Powell, Microscopic
            41.    "k_est10"         - L-BFGS-B, Macroscopic
            42.    "k_est11"         - L-BFGS-B, Microscopic

            Symbolic/Analytical expression of species (refer to 10.1.1)

            43.    "Analyt"          - Pure Symbolic :f(t,xo,k)
            44.    "Analyt-ftx"      - Semi-Symbolic :f(t,xo)
            45.    "SAnalyt"         - Semi-Symbolic :f(t)
            46.    "SAnalyt-ftk"     - Semi-Symbolic :f(t,k)
            47.    "Analyt2"         - Creates commands for wxmaxima
    """
    # global FILE_NAME, ITUPS
    with open(FILE_NAME["topology"], "w") as ffvar:
        ffvar.write(FILE_NAME['last_open'].get("0.0", "end"))

    path = Path(FILE_NAME["topology"])
    s_s = str(FILE_NAME["topology"]).split("/")
    s_s = s_s[-1] if len(s_s) > 1 else ""
    name = str(path.parent) + "/" + s_s + "_" + \
        datetime.now().strftime("%Y%m%d_%H%M%S")
    par = gui.Toplevel()
    par.resizable(False, False)
    par.wm_title("Parameter setting")
    opts = [
        "File name",
        "Number of iteration :",
        "File Units? :",
        "Volume (L) :",
        "end time (tend) :",
        "tau-scaler",
        "Normalized",
        "logx",
        "logy",
        "method",
        "tsteps",
        "mix_plot",
        "save",
        "out fname",
        "show plot",
        "time label",
        "Cini range",
        "K-range",
        "mult proc",
        "Implicit"
    ]
    defs = [
        FILE_NAME["topology"], 1, gui.StringVar(), 1.0, 100, 1.5, False, False,
        False, method, 1000, True, True, name, True, "time (sec)", "", "",
        False, False]
    defs[2].set('molecules')

    topfile = open(FILE_NAME["topology"], "r")
    for row in topfile:
        if row[0] == "#":
            g_g = row.split(",")[1:]
            for xvar in g_g:
                x_x = [g.strip() for g in xvar.split("=")]
                if x_x[0] == "Volume":
                    defs[3] = x_x[1]
                elif x_x[0] == "tend":
                    defs[4] = x_x[1]
                elif x_x[0] == "FileUnit":
                    defs[2].set(x_x[1])
                elif x_x[0] == "logx":
                    defs[7] = x_x[1]
                elif x_x[0] == "Normalized":
                    defs[6] = x_x[1]
                elif x_x[0] == "steps":
                    defs[10] = x_x[1]

    oplen = len(opts)
    entry_lst = [gui.Label(par, text=opts[i], fg="blue") for i in range(oplen)]
    for i in range(10):
        entry_lst[i].grid(row=i, column=0, sticky=gui.W, pady=2)
    for i in range(10):
        entry_lst[10 + i].grid(row=i, column=2, sticky=gui.W, pady=2)\

    entry_list = [gui.Entry(par, bd=5) if i != 2 else gui.OptionMenu(
        par, defs[2], 'molecules', 'molar', 'moles') for i in range(oplen)]
    for i in range(10):
        entry_list[i].grid(row=i, column=1, sticky=gui.W, pady=2)
    for i in range(10):
        entry_list[10 + i].grid(row=i, column=3, sticky=gui.W, pady=2)
    for i in range(oplen):
        if i != 2:
            entry_list[i].insert(gui.END, str(defs[i]))
    entry_list[9].configure(state="disable")
    entry_list[2].config(width=14)
    if method == "ODE":
        entry_list[5].configure(state="disable")
    if method == "ODE2":
        entry_list[5].configure(state="disable")
    elif method == "Gillespie_":
        entry_list[5].configure(state="disable")
    elif method == "CLE":
        entry_list[5].delete(0, gui.END)
        entry_list[5].insert(gui.END, str(10))

    but_b1 = ttk.Button(par, text="RUN",
                        command=lambda: mrun_propagation(par, entry_list,
                                                         defs))
    but_b1.grid(row=oplen, column=0, sticky=gui.W, pady=2)
    if method in ["k_est1", "k_est2", "k_est3", "k_est4", "k_est6", "k_est7",
                  "k_est8", "k_est9", "k_est10", "k_est11", "LNA2", "LNA3",
                  "LNA-vs", "LNA-ks", "LNA-xo", "NetLoc1", "NetLoc2", "Analyt",
                  "SAnalyt-ftk", "SAnalyt", "Analyt-ftx", "Analyt2",
                  "topoTosbml", "topoTosbml2", "topoTosbml3"]:
        but_b1.invoke()


if __name__ == "__main__":

    MENUBUT1 = gui.Menubutton(
        FRAME, text=" File/Model ", activebackground="#f2f20d",
        activeforeground="red", bg="#00cc00",
        fg="white" if PLATFORM.lower() != "darwin" else "green")
    MENUBUT1.menu = gui.Menu(MENUBUT1, tearoff=1)
    MENUBUT1["menu"] = MENUBUT1.menu
    LOADMENU = gui.Menu(FRAME, tearoff=1)
    LOADMENU.add_command(
        label="Topology/File", command=lambda: load_data(ITUPS),
        background="white", foreground="Blue")
    LOADMENU.add_command(
        label="Trajectory file",
        command=tload_data2, background="white", foreground="Blue")
    LOADMENU.add_command(
        label="Traj. w/ plot", command=lambda: tload_data2(True),
        background="white", foreground="Blue")
    LOADMENU.add_command(
        label="Image of plot", command=load_image,
        background="white", foreground="Blue")
    LOADMENU.add_command(
        label="Image w/ data", command=lambda: load_image(True),
        background="white", foreground="Blue")
    LOADMENU.add_command(
        label="Current folder",
        command=lambda: show_file_dir(FILE_NAME["current_folder"]),
        background="white", foreground="Blue")
    MENUBUT1.menu.add_cascade(label="Open", menu=LOADMENU)
    NEWFMENU = gui.Menu(FRAME, tearoff=1)
    NEWFMENU.add_command(label="Blank File", command=lambda: create_file(
        ITUPS, 0), background="white", foreground="Blue")
    NEWFMENU.add_command(label="Topo File", command=lambda: create_file(
        ITUPS, 1), background="white", foreground="Blue")
    NEWFMENU.add_command(label="ODE File", command=lambda: create_file(
        ITUPS, 2), background="white", foreground="Blue")
    MENUBUT1.menu.add_cascade(label="New File", menu=NEWFMENU)
    MENUBUT1.menu.add_command(label="Save File", command=lambda: save_file())
    MENUBUT1.menu.add_command(
        label="Run File.py", command=lambda: runpy_file())
    MENUBUT1.menu.add_command(label="Run SSL", command=lambda: run_ssl())
    CONVMENU = gui.Menu(FRAME, tearoff=1)
    CONVMENU.add_command(label="SBML to Topo", command=lambda: sbml_to_topo2(
        globals2.TO_CONVERT, ITUPS), background="white", foreground="Blue")
    CONVMENU.add_command(label="ODE to Topo", command=lambda: extract_ode(
        ITUPS), background="white", foreground="Blue")
    TOPSBML = gui.Menu(FRAME, tearoff=1)
    TOPSBML.add_command(label="molecules", command=lambda: param_set(
        "topoTosbml"), background="white", foreground="Blue")
    TOPSBML.add_command(label="molar", command=lambda: param_set(
        "topoTosbml2"), background="white", foreground="Blue")
    TOPSBML.add_command(label="no unit", command=lambda: param_set(
        "topoTosbml3"), background="white", foreground="Blue")
    CONVMENU.add_cascade(label="Topo to SBML", menu=TOPSBML)
    MENUBUT1.menu.add_cascade(label="Convert model", menu=CONVMENU)
    PARESMENU = gui.Menu(FRAME, tearoff=1)
    PARESMENU.add_command(
        label="Nelder-Mead (NM), Macroscopic",
        command=lambda: param_set("k_est6"),
        background="white", foreground="Blue")
    PARESMENU.add_command(
        label="Nelder-Mead (NM), Microscopic",
        command=lambda: param_set("k_est7"), background="white",
        foreground="Blue")
    PARESMENU.add_command(
        label="Powell, Macroscopic", command=lambda: param_set("k_est8"),
        background="white", foreground="Blue")
    PARESMENU.add_command(
        label="Powell, Microscopic", command=lambda: param_set("k_est9"),
        background="white", foreground="Blue")
    PARESMENU.add_command(
        label="L-BFGS-B, Macroscopic",
        command=lambda: param_set("k_est10"), background="white",
        foreground="Blue")
    PARESMENU.add_command(
        label="L-BFGS-B, Microscopic",
        command=lambda: param_set("k_est11"), background="white",
        foreground="Blue")
    PARESMENU.add_command(
        label="NM-Diff. Evol., Macroscopic",
        command=lambda: param_set("k_est3"), background="white",
        foreground="Blue")
    PARESMENU.add_command(
        label="NM-Diff. Evol., Microscopic",
        command=lambda: param_set("k_est4"), background="white",
        foreground="Blue")
    PARESMENU.add_command(
        label="Parameter slider/scanner",
        command=lambda: param_set("k_est5"), background="white",
        foreground="Blue")
    PARESMENU.add_command(
        label="MCEM, Macroscopic",
        command=lambda: param_set("k_est1"), background="white",
        foreground="Blue")
    PARESMENU.add_command(
        label="MCEM, Microscopic",
        command=lambda: param_set("k_est2"), background="white",
        foreground="Blue")
    MENUBUT1.menu.add_cascade(label="Estimate Params", menu=PARESMENU)
    MENUBUT1.place(x=2, y=5)

    MENUBUT2 = gui.Menubutton(
        FRAME, text="Propagation", activebackground="#f2f20d",
        activeforeground="red", bg="#00cc00",
        fg="white" if PLATFORM.lower() != "darwin" else "green")
    MENUBUT2.menu = gui.Menu(MENUBUT2, tearoff=1)
    MENUBUT2["menu"] = MENUBUT2.menu
    ANALMENU = gui.Menu(FRAME, tearoff=1)
    ANALMENU.add_command(
        label="Pure Symbolic :f(t,xo,k)",
        command=lambda: param_set("Analyt"), background="white",
        foreground="Blue")
    ANALMENU.add_command(
        label="Semi-Symbolic :f(t)", command=lambda: param_set("SAnalyt"),
        background="white",
        foreground="Blue")
    ANALMENU.add_command(
        label="Semi-Symbolic :f(t,xo)",
        command=lambda: param_set("Analyt-ftx"), background="white",
        foreground="Blue")
    ANALMENU.add_command(label="Semi-Symbolic :f(t,k)", command=lambda: param_set(
        "SAnalyt-ftk"), background="white", foreground="Blue")
    ANALMENU.add_command(label="For wxmaxima", command=lambda: param_set(
        "Analyt2"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="Analytical soln.", menu=ANALMENU)

    ODEMENU = gui.Menu(FRAME, tearoff=1)
    ODEMENU.add_command(label="Molecules(micro)", command=lambda: param_set(
        "ODE-1"), background="white", foreground="Blue")
    ODEMENU.add_command(label="Molar(macro)", command=lambda: param_set(
        "ODE-2"), background="white", foreground="Blue")
    ODEMENU.add_command(label="Mole(macro)", command=lambda: param_set(
        "ODE-3"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="ODE int", menu=ODEMENU)

    RUNGEK4 = gui.Menu(FRAME, tearoff=1)
    RUNGEK4.add_command(label="Molecules(micro)", command=lambda: param_set(
        "rk4-1"), background="white", foreground="Blue")
    RUNGEK4.add_command(label="Molar(macro)", command=lambda: param_set(
        "rk4-2"), background="white", foreground="Blue")
    RUNGEK4.add_command(label="Mole(macro)", command=lambda: param_set(
        "rk4-3"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="RK4-fix-interval", menu=RUNGEK4)

    RUNGEK4A = gui.Menu(FRAME, tearoff=1)
    RUNGEK4A.add_command(label="Molecules(micro)", command=lambda: param_set(
        "rk4-1a"), background="white", foreground="Blue")
    RUNGEK4A.add_command(label="Molar(macro)", command=lambda: param_set(
        "rk4-2a"), background="white", foreground="Blue")
    RUNGEK4A.add_command(label="Mole(macro)", command=lambda: param_set(
        "rk4-3a"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="RK4-tau-adaptive", menu=RUNGEK4A)

    EULRTAU = gui.Menu(FRAME, tearoff=1)
    EULRTAU.add_command(label="Molecules(micro)", command=lambda: param_set(
        "Euler-1"), background="white", foreground="Blue")
    EULRTAU.add_command(label="Molar(macro)", command=lambda: param_set(
        "Euler-2"), background="white", foreground="Blue")
    EULRTAU.add_command(label="Mole(macro)", command=lambda: param_set(
        "Euler-3"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="Euler (tau-adaptive-1)", menu=EULRTAU)

    EULRTAU2 = gui.Menu(FRAME, tearoff=1)
    EULRTAU2.add_command(label="Molecules(micro)", command=lambda: param_set(
        "Euler2-1"), background="white", foreground="Blue")
    EULRTAU2.add_command(label="Molar(macro)", command=lambda: param_set(
        "Euler2-2"), background="white", foreground="Blue")
    EULRTAU2.add_command(label="Mole(macro)", command=lambda: param_set(
        "Euler2-3"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="Euler (tau-adaptive-2)", menu=EULRTAU2)

    CLETAUA = gui.Menu(FRAME, tearoff=1)
    CLETAUA.add_command(label="Molecules(micro)", command=lambda: param_set(
        "CLE"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="CLE (tau-adaptive)", menu=CLETAUA)

    CLETAUA2 = gui.Menu(FRAME, tearoff=1)
    CLETAUA2.add_command(label="Molecules(micro)", command=lambda: param_set(
        "CLE2"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="CLE (cle-fixIntvl)", menu=CLETAUA2)

    TAULMENU = gui.Menu(FRAME, tearoff=1)
    TAULMENU.add_command(label="Tau-leapingV1-micro", command=lambda: param_set(
        "Tau-leaping"), background="white", foreground="Blue")
    TAULMENU.add_command(label="Tau-leapingV2-micro", command=lambda: param_set(
        "Tau-leaping2"), background="white", foreground="Blue")
    TAULMENU.add_command(label="Sim-TauLeap-micro", command=lambda: param_set(
        "Sim-TauLeap"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="Tau-leaping-micro", menu=TAULMENU)

    LNAMENU = gui.Menu(FRAME, tearoff=1)
    LNAMENU.add_command(label="COV-time-dependent, Macroscopic",
                        command=lambda: param_set("LNA(t)"), background="white", foreground="Blue")
    LNAMENU.add_command(label="FF-time-dependent, Macroscopic",
                        command=lambda: param_set("LNA2(t)"), background="white", foreground="Blue")
    LNAMENU.add_command(label="Numeric, values", command=lambda: param_set(
        "LNA"), background="white", foreground="Blue")
    LNAMENU.add_command(label="Symbolic, Microscopic", command=lambda: param_set(
        "LNA2"), background="white", foreground="Blue")
    LNAMENU.add_command(label="Symbolic, Macroscopic", command=lambda: param_set(
        "LNA3"), background="white", foreground="Blue")
    LNAMENU.add_command(label="Symbolic, f(xo), Macroscopic", command=lambda: param_set(
        "LNA-xo"), background="white", foreground="Blue")
    LNAMENU.add_command(label="Symbolic, f(ks), Macroscopic", command=lambda: param_set(
        "LNA-ks"), background="white", foreground="Blue")
    LNAMENU.add_command(label="Symbolic, values, Macroscopic", command=lambda: param_set(
        "LNA-vs"), background="white", foreground="Blue")
    MENUBUT2.menu.add_cascade(label="Linear Noise Appx.", menu=LNAMENU)

    GILMENU = gui.Menu(FRAME, tearoff=1)
    GILMENU.add_command(label="Direct method", command=lambda: param_set(
        "Gillespie_"), background="white", foreground="Blue")
    # GILMENU.add_command ( label="First Rxn Method",command=lambda:
    # print("Not implemented yet"),background="white",foreground="Blue"  )
    # GILMENU.add_command ( label="Next Rxn Method",command=lambda:
    # print("Not implemented yet"),background="white",foreground="Blue"  )
    # GILMENU.add_command ( label="Optimized Direct",command=lambda:
    # print("Not implemented yet"),background="white",foreground="Blue"  )
    MENUBUT2.menu.add_cascade(label="Gillespie", menu=GILMENU)
    MENUBUT2.place(x=95, y=5)

    MENUBUT3 = gui.Menubutton(
        FRAME, text="    Analysis    ", activebackground="#f2f20d",
        activeforeground="red", bg="#00cc00",
        fg="white" if PLATFORM.lower() != "darwin" else "green")
    MENUBUT3.menu = gui.Menu(MENUBUT3, tearoff=1)
    MENUBUT3["menu"] = MENUBUT3.menu
    MENUBUT3.menu.add_command(
        label="Covariance", command=lambda: analysis_case("cov", ITUPS),
        background="white", foreground="Blue")
    MENUBUT3.menu.add_command(
        label="Fano Factor", command=lambda: analysis_case("fanoF", ITUPS),
        background="white", foreground="Blue")
    MENUBUT3.menu.add_command(
        label="Cross correlation",
        command=lambda: analysis_case("corr", ITUPS), background="white",
        foreground="Blue")
    MENUBUT3.menu.add_command(
        label="Probability density",
        command=lambda: analysis_case("pdens1", ITUPS), background="white",
        foreground="Blue")
    MENUBUT3.menu.add_command(
        label="Freq. Dist w/r to t",
        command=lambda: analysis_case("pdens2", ITUPS), background="white",
        foreground="Blue")
    MENUBUT3.menu.add_command(
        label="Hist. slice of time",
        command=lambda: analysis_case("pdens3", ITUPS), background="white",
        foreground="Blue")
    MENUBUT3.menu.add_command(
        label="Average of traj.",
        command=lambda: analysis_case("avetrj", ITUPS), background="white",
        foreground="Blue")
    MENUBUT3.menu.add_command(
        label="Phase portrait",
        command=lambda: analysis_case("phaseP", ITUPS), background="white",
        foreground="Blue")
    MENUBUT3.menu.add_command(
        label="Plot Data", command=lambda: analysis_case("plotD", ITUPS),
        background="white", foreground="Blue")
    NETLMENU = gui.Menu(FRAME, tearoff=1)
    NETLMENU.add_command(
        label="Symbolic, Macroscopic",
        command=lambda: param_set("NetLoc1"), background="white",
        foreground="Blue")
    NETLMENU.add_command(
        label="Numeric, Macroscopic", command=lambda: param_set("NetLoc2"),
        background="white", foreground="Blue")
    MENUBUT3.menu.add_cascade(label="Network Localization", menu=NETLMENU)
    MENUBUT3.place(x=189, y=5)

    FRAME1 = gui.Frame(FRAME, height=435, width=972,
                       bg='#8c8c8c', borderwidth=2)
    FRAME1.place(x=0, y=35, relheight=0.93, relwidth=1.0)
    ITUPS = prepare_frame_for_plot(FRAME1, 972, 435)
    load_data2(True, [test_data2.data, test_data2.label])
    TOP.mainloop()



"""

                   This module is the lna_approx2 module

This module handles symbolic linear noise approximation.

The following are the list of function in this module;

1. subs2
2. lna_symbolic2

"""

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

from sympy import simplify, Matrix, solve, Symbol, flatten, nonlinsolve, \
    collect, factor
from BioSANS2020.gui_functs.scrollable_text import prepare_scroll_text
from BioSANS2020.propagation.propensity \
    import propensity_vec, propensity_vec_molar


def subs2(zvar, cval):
    """This function helps in the substitution of value to sympy Symbols

    Args:
        zvar (Symbol): sympy expression
        cval (dict): dictionary of values

    Returns:
        (Symbol): substituted expression
    """
    for xvar in cval:
        zvar = zvar.subs(xvar, cval[xvar])
    return zvar


def lna_symbolic2(sp_comp, ks_dict, conc, r_dict, p_dict, stch_var,
                  items=None, molar=False, mode=None):
    """This function facilitates in the symbolic LNA computation.

    Args:
        sp_comp (dict): dictionary of appearance or position of species
            or component in the reaction tag of BioSANS topology file.

            For example;

                #REACTIONS
                A => B, kf1
                B => C, kf2

            The value of sp_comp is

                sp_comp = {'A': {0}, 'B': {0, 1}, 'C': {1}}

                A appears in first reaction with index 0
                B appears in first and second reaction with index 0, 1
                C appears in second reaction with index 1
        ks_dict (dict): dictionary of rate constant that appears in each
            reactions.

            For example;

                #REACTIONS
                A => B , 0.3        # first reaction
                B <=> C, 0.1, 0.2   # second reaction

            The value of ks_dict is

                ks_dict = {
                    0 : [0.3],      # first reaction
                    1 : [0.1, 0.2]  # second reaction
                }
        conc ([type]): [description]
        r_dict (dict): dictionary of reactant stoichiometry. For example

            r_dict = {
                0: {'A': 1},  # first reaction, coefficient of A is 1
                1: {'B': 1}   # second reaction, coefficient of B is 1
            }
        p_dict (dict): dictionary of product stoichiometry. For example

            p_dict = {
                0: {'B': 1},  # first reaction, coefficient of B is 1
                1: {'C': 1}   # second reaction, coefficient of C is 1
            }
        stch_var (numpy.ndarray): stoichiometric matrix. For example

            v_stoich = np.array([
                [   -1,           0   ]            # species A
                [    1,          -1   ]            # species B
                [    0,           1   ]            # species C
                  #1st rxn    2nd rxn
            ])
        items (list): list of [canvas, scroll_x, scroll_y]
        molar (bool, optional): If True, the units for any amount is in
            molar. Propensity will be macroscopic. Defaults to False.
        mode (str, optional): method keywords : Numeric, fofks, fofCo

    Returns:
        list: [0, 0] - not used
    """

    c_s = {}
    c_so = {}
    equivals = []
    equi_co = []
    equi_ks = []

    for xvar in sp_comp:
        c_s[xvar] = Symbol(xvar, real=True, negative=False)
        c_so[xvar] = Symbol(xvar + 'o', real=True, negative=False) * \
            (0 if conc[xvar] == 0 else 1)
        equivals.append((c_so[xvar], conc[xvar]))
        equi_co.append((c_so[xvar], conc[xvar]))

    kcs = []
    for i, _ in enumerate(ks_dict):
        row = []
        if len(ks_dict[i]) == 1:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, real=True, negative=False))
            equivals.append((row[0], ks_dict[i][0]))
            equi_ks.append((row[0], ks_dict[i][0]))
        else:
            key = 'kf' + str(i + 1)
            row.append(Symbol(key, real=True, negative=False))
            equivals.append((row[0], ks_dict[i][0]))
            equi_ks.append((row[0], ks_dict[i][0]))
            key = 'kb' + str(i + 1)
            row.append(Symbol(key, real=True, negative=False))
            equivals.append((row[1], ks_dict[i][1]))
            equi_ks.append((row[1], ks_dict[i][1]))
        kcs.append(row)

    if not molar:
        prop_flux = Matrix(propensity_vec(kcs, c_s, r_dict, p_dict))
    else:
        prop_flux = Matrix(propensity_vec_molar(kcs, c_s, r_dict, p_dict))

    if mode == "Numeric":
        prop_flux = prop_flux.subs(equivals)
    elif mode == "fofks":
        prop_flux = prop_flux.subs(equi_co)
    elif mode == "fofCo":
        prop_flux = prop_flux.subs(equi_ks)

    stch_var = Matrix(stch_var)
    # c_s might have change after call
    for xvar in sp_comp:
        c_s[xvar] = Symbol(xvar, real=True, negative=False)
    slabels = [xvar for xvar in c_s]

    s_s = []
    n_z = []
    for row in range(stch_var.shape[0]):
        if sum(abs(stch_var[row, :])) != 0 and slabels[row][0] != "-":
            s_s.append(list(stch_var[row, :]))
            n_z.append(row)
    s_s = Matrix(s_s)

    da_dt = s_s * prop_flux
    # print(da_dt)
    ccs = [c_s[xvar] for xvar in c_s]
    j_s = [ccs[xvar] for xvar in n_z]
    a_jac = da_dt.jacobian(j_s)
    diag_prof = [[0] * len(prop_flux) for xvar in range(len(prop_flux))]
    for i, _ in enumerate(prop_flux):
        diag_prof[i][i] = diag_prof[i][i] + prop_flux[i]
    diag_prof = Matrix(diag_prof)

    bbt = s_s * diag_prof * s_s.T

    cov = []
    for i in n_z:
        row = []
        for j in n_z:
            key = "C" + str(min(i + 1, j + 1)) + "_" + str(max(i + 1, j + 1))
            if i == j:
                row.append(Symbol(key, real=True, negative=False))
            else:
                row.append(Symbol(key, real=True))
        cov.append(row)
    cov = Matrix(cov)
    # print(s_s)
    same = {}
    for i in range(s_s.shape[0] - 1):
        for j in range(i + 1, s_s.shape[0]):
            if s_s[i, :] == s_s[j, :]:
                same[(i, j)] = 1
            elif s_s[i, :] == -s_s[j, :]:
                same[(i, j)] = -1
            else:
                pass

    dcov_dt = a_jac * cov + cov * a_jac.T + bbt

    x_s = list(dict.fromkeys(flatten(cov)))

    if items:
        text = prepare_scroll_text(items)

        def ffprint(xvar):
            return text.insert('insert', " ".join([str(y) for y in xvar]))
    else:

        def ffprint(xvar):
            return print(" ".join([str(y) for y in xvar]), end="")

    sps = [xvar for xvar in sp_comp]
    csps = {}
    for i in n_z:
        for j in n_z:
            if j >= i:
                key = "C" + str(min(i + 1, j + 1)) + "_" + \
                    str(max(i + 1, j + 1))
                ffprint([key, " = ", str(sps[i]) + "_" + str(sps[j]), "\n"])
                csps[key] = sps[i] + "_" + sps[j]

    eqs = []
    ffprint(["\nEquations to solve\n\n"])
    for i in range(len(n_z)):
        for j in range(i, len(n_z)):
            if dcov_dt[i, j] != 0:
                row = collect(dcov_dt[i, j], x_s)
                eqs.append(row)
                ffprint([row, "\n"])

    for e_n in same:
        i, j = e_n
        for k in range(j, s_s.shape[0]):
            eqs.append(cov[j, k] - same[e_n] * cov[i, k])

    post_proc = False
    multiple_cval = False

    val = solve(da_dt, j_s)

    cval = {}
    if val:
        for xvar in j_s:
            key = xvar
            if key in val:
                cval[key] = val[key]
            else:
                post_proc = True
    else:
        post_proc = True

    if post_proc:
        ffprint(
            ["\nSteady state concentrations not properly calculated",
             "due to incomplete constraint\n\n"])
        ffprint(["\n", da_dt, "\n"])
        ffprint(["\nAttempting the use of boundary condition", "\n"])

        tosum = set()
        sets = [tosum]
        used = set()
        for ihvar, _ in enumerate(r_dict):
            used2 = set()
            if len(r_dict[ihvar]) == 1:
                for svar in sets:
                    key = list(r_dict[ihvar].keys())[0]
                    if key not in used and r_dict[ihvar][key] != 0:
                        svar.add(key)
                        used2.add(key)
            else:
                for j, _ in enumerate(sets):
                    svar = sets[j]
                    sets.append(svar.copy())
                    key = list(r_dict[ihvar].keys())[0]
                    if key not in used and r_dict[ihvar][key] != 0:
                        svar.add(key)
                        used2.add(key)
                    key = list(r_dict[ihvar].keys())[1]
                    if key not in used and r_dict[ihvar][key] != 0:
                        sets[-1].add(key)
                        used2.add(key)

            if len(p_dict[ihvar]) == 1:
                for svar in sets:
                    key = list(p_dict[ihvar].keys())[0]
                    if key not in used and p_dict[ihvar][key] != 0:
                        svar.add(key)
                        used2.add(key)
            else:
                for j, _ in enumerate(sets):
                    svar = sets[j]
                    sets.append(svar.copy())
                    key = list(p_dict[ihvar].keys())[0]
                    if key not in used and p_dict[ihvar][key] != 0:
                        svar.add(key)
                        used2.add(key)
                    key = list(p_dict[ihvar].keys())[1]
                    if key not in used and p_dict[ihvar][key] != 0:
                        sets[-1].add(key)
                        used2.add(key)
            for z_s in used2:
                used.add(z_s)

        f_e = [da_dt]
        for svar in sets:
            f_e.append(sum([c_s[xvar] - c_so[xvar] for xvar in svar]))

        if mode == "Numeric":
            f_e = [entry.subs(equivals) for entry in f_e]
        elif mode == "fofks":
            f_e = [entry.subs(equi_co) for entry in f_e]
        elif mode == "fofCo":
            f_e = [entry.subs(equi_ks) for entry in f_e]

        val2 = solve(f_e, {xvar for xvar in j_s})
        cval = {}
        for xvar in j_s:
            key = xvar
            if key in val2:
                cval[key] = val2[key]
            else:
                multiple_cval = True
                break
        # print(f_e)

    ffprint(["\nUsing Algebraic manipulation of AC + CA.T + BT = 0\n"])
    if not multiple_cval:
        print(1)
        ffprint(["\nSteady state concentrations\n\n"])
        for xvar in cval:
            ffprint([xvar, " = ", cval[xvar], "\n\n"])
        dcov_dt = simplify(subs2(dcov_dt, cval))
        eqs = []
        ffprint(["\nEquations to solve\n\n"])
        for i in range(len(n_z)):
            for j in range(i, len(n_z)):
                if dcov_dt[i, j] != 0:
                    row = collect(dcov_dt[i, j], x_s)
                    eqs.append(row)
                    ffprint([row, "\n\n"])

        for e_n in same:
            i, j = e_n
            for k in range(j, s_s.shape[0]):
                eqs.append(cov[j, k] - same[e_n] * cov[i, k])

        sol = solve(eqs, x_s)

        if not sol:
            # LA = Matrix(eqs)
            # LHS = LA.jacobian(x_s)
            # RHS = LA-LHS*Matrix(x_s)
            # sol = simplify(-(LHS**-1)*RHS)
            sol = list(nonlinsolve(eqs, x_s))
            sol = {x_s[xvar]: sol[xvar] for xvar in range(len(x_s))}
        ffprint(["\nCovariance\n\n"])
        for xvar in sol:
            c_h = factor(simplify(subs2(sol[xvar], cval)))
            # c_h = sol[xvar]
            ffprint(["Cov(" + str(csps[str(xvar)]) + ")", " = ", c_h, "\n\n"])
    else:
        print(2)
        ffprint(["\nMultiple solutions detected"])
        valset = 0
        for val in val2:
            ffprint(["\nValues ", valset, "\n"])
            cval = {}
            for xvar in j_s:
                key = xvar
                if key in val:
                    cval[key] = val[key]
            ffprint(["\nSteady state concentrations\n\n"])
            for xvar in cval:
                ffprint([xvar, " = ", cval[xvar], "\n\n"])
            dcov_dt = simplify(subs2(dcov_dt, cval))
            eqs = []
            ffprint(["\nEquations to solve\n\n"])
            for i in range(len(n_z)):
                for j in range(i, len(n_z)):
                    if dcov_dt[i, j] != 0:
                        row = collect(dcov_dt[i, j], x_s)
                        eqs.append(row)
                        ffprint([row, "\n\n"])

            for e_n in same:
                i, j = e_n
                for k in range(j, s_s.shape[0]):
                    eqs.append(cov[i, k] - same[e_n] * cov[j, k])

            sol = solve(eqs, x_s)
            if not sol:
                sol = list(nonlinsolve(eqs, x_s))
                sol = {x_s[xvar]: sol[0][xvar] for xvar in range(len(x_s))}
            ffprint(["\nCovariance\n\n"])
            fact = 1
            for xvar in sol:
                uvar, wvar = str(xvar).split("_")
                c_h = factor(simplify(subs2(sol[xvar], cval))).subs(equivals)
                if uvar == 'C' + wvar and c_h < 0:
                    fact = -1
                    break

            for xvar in sol:
                c_h = factor(simplify(subs2(sol[xvar], cval)))
                # c_h = sol[xvar]
                ffprint(["Cov(" + str(csps[str(xvar)]) + ")",
                         " = ", c_h * fact, "\n\n"])
            valset = valset + 1

    return [0, 0]



"""

                 This module is the my_mcem module

This module is a simple implementation of monte-carlo expectation maximi
zation. The likelihood  function here is the joint probability distribu-
tion of the  error or  the difference between the true value and estima-
ted value. The error on each parameter estimate  is assumed to  be inde-
pendent and follows a normal  distribution. A uniform prior was set. The
log-likelihood  of  this function becomes the negative of the sum of the
squared  error  between  the  estimated value and true value. Parameters
were drawn  from  a log-normal distribution. In our  implementation, the
posterior  probability  is  calculated at each sampling  stage.  In  the
maximization step, the  ratio of posterior  probability between consecu-
tive draws (or the exponential of the difference between consecutive log
-likelihood) are  used to decide  whether to accept or reject the latest
values of the parameters based on a uniform random variable. After seve-
ral samplings  (decided programmatically in the algorithm), the mean and
standard deviation of the parameters  from all  the accepted  values are
calculated  and  used as the mean  and standard  deviation  for the next
sampling  stage.  This serves as the expectation step in our implementa-
tion because we  do not have a  close form for the parameters. The cycle
is repeated until  the calculated  mean and  standard deviations of each
parameters are no longer changing or  until  the maximum number of steps
is reached.

The following are the functions inside this module

1. log_likelihood
2. cost_value
3. exptn_maxtn
4. run_mcem


"""

# Metropolis Hasting Algorithm + Expectation Maximization Algorithm
# Personally coded by Erickson Fajiculay

# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

# import time
# import math
# from scipy import linalg
# from scipy.optimize import fsolve
# import matplotlib.pyplot as plt
# from scipy.integrate import odeint
# import random as rn

import warnings
import numpy as np
from BioSANS2020.myglobal import proc_global


warnings.filterwarnings('ignore')


def log_likelihood(ks_var, custom_function, args=None):
    """This function evaluates the loglikelihood based on the definitons
    provided in the custom_function. Here, it is the negative of the sum
    of squared errors between true and estimated value.

    Args:
        ks_var (list): list of parameter values
        custom_function (function): user/program defined objective/error
            function  to  optimized  my expectation  maximization  or EM
        args (tuple): Tuple  of  (data,  conc, tvar,  sp_comp,  ks_dict,
            r_dict,  p_dict,  v_stoich,  c_miss,  k_miss, molar, rfile).
            Defaults  to None. See param_estimate module for  the proper
            definition of variables.

    Returns:
        float: numeric value of custom_function(ks_var, args)
    """
    return custom_function(ks_var, args)


# def uniform_prior():
#     return 1.0


def cost_value(ks_var, custom_function, args=None):
    """This function  evaluates  the cost value or SSE of the definitons
    provided in the custom_function.

    Args:
        ks_var (list): list of parameter values
        custom_function (function): user/program defined objective/error
            function  to  optimized  my  expectation  maximization or EM
        args (tuple): Tuple  of  (data,  conc, tvar,  sp_comp,  ks_dict,
            r_dict,  p_dict,  v_stoich,  c_miss,  k_miss, molar, rfile).
            Defaults  to None. See param_estimate module for  the proper
            definition of variables.

    Returns:
        float: numeric value of custom_function(ks_var, args)
    """
    return abs(log_likelihood(ks_var, custom_function, args))


def exptn_maxtn(lst, seed_var, maxiter=50, inner_loop=1000, lenks=3,
                positive_only=False, likelihood=None, args=None, thr=1.0e-10):
    """This  function  performs  the  expectation  step and maximization
    steps  bounded by  maxiter and inner_loop  variables. The process is
    similar to the metrololis  hasting algorithm. There is a random walk
    performed as the  posterior probability is maximized. The parameters
    are updated in the randomwalk based on sampled values.

    Args:
        lst (multiprocessing.managers.ListProxy):
            multiprocessing.Manager() from proc_global module. This help
            in halting simulation when one of the chains already achived
            the defined tolerance.
        seed_var (float): random  seed  value picked  at random for each
            trajectory. They have been sampled from the calling program.
        maxiter (int, optional):  maximum  number  of expectation steps.
            Defaults to 50. This is the outer loop.
        inner_loop (int, optional): maximum number of maximization step.
            Defaults to 1000.
        lenks (int, optional): numbers  unknown  parameters to estimate.
        positive_only (bool, optional): constraint the parameter estima-
            tion to choose positive values. Defaults to False.
        likelihood (function, optional): user/program defined objective/
            error function  to  optimized
        args (tuple): Tuple  of  (data,  conc, tvar,  sp_comp,  ks_dict,
            r_dict,  p_dict,  v_stoich,  c_miss,  k_miss, molar, rfile).
            Defaults  to None. See param_estimate module for  the proper
            definition of variables.
        thr (function, optional): error tolerance. Defaults to 1.0e-10.

    Returns:
        tuple: (list of estimated parameter, minimum error)
    """

    np.random.seed(int(seed_var * 100))
    sd_var = np.random.uniform(0, 1)

    stds = [0] * lenks
    for i in range(lenks):
        if positive_only:
            stds[i] = abs(np.random.normal(sd_var, 0.01))
        else:
            stds[i] = abs(np.random.normal(sd_var, 100))

    kso = [0] * lenks
    for i in range(lenks):
        if positive_only:
            kso[i] = np.random.lognormal(0, stds[i])
        else:
            kso[i] = np.random.normal(0, stds[i])

    valo = [0] * lenks
    valp = [0] * lenks
    for i in range(lenks):
        valo[i] = 1
        valp[i] = 2

    ks_var = []
    for i in range(lenks):
        ks_var.append([])

    rs_var = []

    vmax = 1.0e+100
    iteration = 0
    test = 1000
    error = 1000
    # sigma = 1

    delta = int(np.ceil(inner_loop / (maxiter - 0.33 * maxiter)))
    inner_loop_orig = inner_loop
    inner_loop = delta
    while iteration < maxiter:
        if lst:
            return (valp, 1e+100)
        kis = []
        for i in range(lenks):
            kis.append([])

        r_big = vmax * log_likelihood(kso, likelihood, args)

        # Maximization step by metropolis hasting algorithm
        for i in range(inner_loop):
            if lst:
                break
            ksp = [0] * lenks
            for j in range(lenks):
                if positive_only:
                    ksp[j] = np.random.lognormal(
                        np.log(kso[j]), max(0, stds[j]))
                else:
                    ksp[j] = np.random.normal(kso[j], stds[j])

            l_big = vmax * log_likelihood(ksp, likelihood, args)

            u_rand = np.random.uniform(0, 1)
            check = np.exp(l_big - r_big)
            if check > u_rand:

                for j in range(lenks):
                    kis[j].append(ksp[j])
                    kso[j] = ksp[j]

                rs_var.append(l_big)
                r_big = l_big
            else:
                for j in range(lenks):
                    kis[j].append(kso[j])

        # Expectatation step by averaging
        for i in range(lenks):
            if len(kis[i]) > 1:
                for k in kis[i]:
                    ks_var[i].append(k)
                if positive_only:
                    stds[i] = np.std(np.log(kis[i]))
                else:
                    stds[i] = np.std(kis[i])
                valo[i] = valp[i]
                valp[i] = np.array(kis[i]).mean()
            else:
                stds[i] = abs(np.random.normal(2 * stds[i], 0.1 * stds[i]))

        test = 0
        for i in range(lenks):
            test = test + abs((valo[i] - valp[i]) / valo[i])

        if test < 1.0e-15:
            break

        iteration = iteration + 1
        inner_loop = min(inner_loop_orig, inner_loop + delta)

    error = min(error, cost_value(valp, likelihood, args))
    if error < thr:
        lst.append(1)  # early stopping of all chains
    return (valp, error)


def run_mcem(chains, n_pars, maxiter=5, inner_loop=5 * 1000,
             positive_only=False, likelihood=None, arg=None,
             r_rand=np.random.uniform(0, 1)):
    """[summary]

    Args:
        chains (integer): number of parallel runs or chains
        n_pars (integer): number of parameters
        maxiter (int, optional):  maximum  number  of expectation steps.
            Defaults to 50. This is the outer loop.
        inner_loop (int, optional): maximum number of maximization step.
            Defaults to 1000.
        positive_only (bool, optional): constraint the parameter estima-
            tion to choose positive values. Defaults to False.
        likelihood (function, optional): user/program defined objective/
            error function  to  optimized
        args (tuple): Tuple  of  (data,  conc, tvar,  sp_comp,  ks_dict,
            r_dict,  p_dict,  v_stoich,  c_miss,  k_miss, molar, rfile).
            Defaults  to None. See param_estimate module for  the proper
            definition of variables.
        r_rand (float, optional): np.random.uniform(0, 1) for seeding

    Returns:
        [type]: [description]
    """
    m_proc = proc_global.mp
    pool = m_proc.Pool(chains)
    np.random.seed(int(r_rand * 100))
    rands = [(xvar + 1) * np.random.uniform(0, 1) for xvar in range(chains)]

    thr = 1.0e-10
    results = [
        pool.apply_async(
            exptn_maxtn, args=(proc_global.LST, rands[ih], maxiter * (ih + 1),
                               inner_loop * (ih + 1), n_pars, positive_only,
                               likelihood, arg, thr)
        ) for ih in range(chains)
    ]

    ffvar = [result.get() for result in results]
    pool.close()

    ffvar = [result.get() for result in results]
    er_list = []
    for xvar in ffvar:
        er_list.append(xvar[1])

    er_min = min(er_list)
    ks_var = []
    for xvar in ffvar:
        if xvar[1] <= er_min:
            ks_var = xvar[0]
    return (ks_var, er_min)



"""
                       This is the draw_figure module

This module controls how plots are drawn inside BioSANS canvas

The following is the list of function for this module:

1. canvas_update_widgets
2. delete_this
3. draw_figure

"""


# import sys
# import os
# sys.path.append(os.path.abspath("BioSANS2020"))

# from tkinter import Frame, Canvas, Checkbutton, IntVar, Button
from tkinter import Frame, Canvas, IntVar, Button
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, \
    NavigationToolbar2Tk as NavigationToolbar2TkAgg
from BioSANS2020.myglobal import mglobals as globals2


def canvas_update_widgets(_, canvas):
    """This function update the canvas upon adding new objects by moving
    objects to ensure they fit nicely. The arrangement is that the last
    object added appear first (latest goes on top).
    Args:
        canvas : the canvas object"""
    # R = canvas._root().winfo_height()
    # root = canvas._root()
    # width = root.winfo_screenwidth()
    # height = root.winfo_screenheight()
    # print(canvas.winfo_children())
    canvas_height = canvas.winfo_height()
    canvas_width = canvas.winfo_width()

    obj_canv = canvas.find_all()
    obj_len = len(obj_canv)
    ind = 0
    for xvar in obj_canv:
        canvas.itemconfig(xvar, height=canvas_height, width=canvas_width - 5)
        # xx1, yy1, xx2, yy2 = canvas.bbox(xvar)
        _, yy1, _, _ = canvas.bbox(xvar)
        canvas.move(xvar, 0, (obj_len - ind) * canvas_height - yy1 + 3)
        ind = ind + 1
    canvas.configure(scrollregion=canvas.bbox("all"))
    return "break"


def delete_this(frame, canvas):
    """This function removes objects from the canvas and reorder objects"""
    canvas.delete(frame)
    canvas_update_widgets(None, canvas)


def draw_figure(items, figure, loc=(0, 0)):
    """This function draws figure to the canvas by first adding a canvas
    object to the canvas, adding a frame to the newly created canvas and
    putting the figure to FigureCanvasTkAgg with frame as the parent.
    Args:
        items : 3 item list of [canvas, scroll_x, scroll_y]
        figure: the figure to be drawn in canvas i.e. plt.gcf() figure
    """
    if items:
        canvas, scroll_x, scroll_y = items
        canva = Canvas(canvas, height=426, width=1030, bg='#ccffcc')
        frame = Frame(canva, height=425, width=1000, borderwidth=0, bd=0)
        frame.pack(side='top')
        figure_canvas = FigureCanvasTkAgg(figure, frame)
        canva.pack(fill="both", expand=True)
        figure_canvas.get_tk_widget().pack(side='top')
        toolbar = NavigationToolbar2TkAgg(figure_canvas, frame)
        toolbar.update()
        figure_canvas._tkcanvas.pack(side="top")
        globals2.INT_VARS.append(IntVar(value=-1))
        # C1 = Checkbutton(canva, text = "", variable = globals2.INT_VARS[-1],
        #                  onvalue = globals2.PLOT_I, offvalue = -1,
        #                  height=2, width = 2,bg="#ccffcc")
        # C1.place(x=1000,y=200)
        wind1 = canva.create_window(2, 2, anchor='nw', window=frame)
        # wind2 = canva.create_window(1020, 216, anchor='ne', window=C1,
        #                             tags="gg")
        fframe = canvas.create_window(
            0, 450 * globals2.PLOT_I, anchor='nw', window=canva)
        buttn = Button(frame, text=" X ", fg='red', highlightcolor='blue',
                       bg='white', height=1, relief='raised',
                       command=lambda: delete_this(fframe, canvas))
        buttn.place(rely=0.0, relx=1.0, x=-15, y=0, anchor="ne")

        globals2.CONTAINER.append([canvas, wind1])
        # canvas.move(wind1,100,100)
        canvas.configure(yscrollcommand=scroll_y.set,
                         xscrollcommand=scroll_x.set)
        canvas.configure(scrollregion=canvas.bbox("all"))
        globals2.PLOT_I = globals2.PLOT_I + 1
        canvas_update_widgets(None, canvas)
        print(loc)
